#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <err.h>

#define MAXSEQ 512
#define MAXNOTE 20

#define PLONKTIME 18
#define CHUNKTIME1 18
#define CHUNKTIME2 8
//#define CHUNKTIME3 4

enum {
	N_UNUSED,
	N_LEAF,
	N_BRANCH
};

struct node {
	uint16_t weight;
	uint16_t origweight;
	uint8_t	type;
	uint8_t parent;
	uint8_t	value;	// if leaf
	uint8_t alt[2];	// if branch
	uint8_t	address;
} node[256];
int nnode = 0;

struct notemap {
	char	*name;
	int	id;
} notemap[MAXNOTE] = {
	{"e-1", 2},
	{"f#1", 3},
	{"g-1", 4},
	{"g#1", 5},
	{"a-1", 6},
	{"a#1", 7},
	{"h-1", 10},
	{"c-2", 11},
	{"c#2", 12},
	{"d-2", 13},
	{"e-2", 14},
	{"f-2", 15},
	{"f#2", 16},
	{"g-2", 17},
	{"g#2", 18},
	{"a-2", 19},
	{"a#2", 20},
	{"h-2", 21},
	{"c-3", 22},
	{"d-3", 23}
};

uint8_t leaves[256];

uint8_t sequence[MAXSEQ];
int nseq = 0;

#define BRANCH(n)(((n) << 1) | 0x01)	/* [0, 127] */
#define DELAY(n) (((n) << 2) | 0x02)	/* [0, 63] */
#define PLONK(n) (((n) << 3) | 0x04)	/* [0, 31] */
#define FADEUP(n)(((n) << 4) | 0x08)	/* [0, 15] */
#define FADEDN(n)(((n) << 4) | 0x00)	/* [0, 15] */

int bitcount = 0, accum = 0;
int queued_0s = 0;

int totalbytes = 0;

void putbit(int v, int enqueue_0s) {
	static int totalbits = 0;

	if(enqueue_0s && !v) {
		queued_0s++;
	} else {
		while(v && queued_0s) {
			putbit(0, 0);
			queued_0s--;
		}
		accum <<= 1;
		accum |= v;
		bitcount++;
		if(bitcount == 8) {
			printf("\t\t.byte\t0x%02x\n", accum);
			accum = 0;
			bitcount = 0;
			totalbytes++;
		}
	}
}

void chain(uint8_t n) {
	if(node[n].parent != 0xff) {
		chain(node[n].parent);
		putbit(node[node[n].parent].alt[1] == n, 1);
	}
}

int lightest() {
	int i, argmin = -1;
	uint16_t min = 0xffff;

	for(i = 0; i < nnode; i++) {
		if(node[i].weight && node[i].weight < min) {
			argmin = i;
			min = node[i].weight;
		}
	}
	return argmin;
}

void fprintaction(FILE *f, uint8_t j) {
	if(j & 1) {
		fprintf(f, "branch(%d)", j >> 1);
	} else {
		j >>= 1;
		if(j & 1) {
			fprintf(f, "delay(%d)", j >> 1);
		} else {
			j >>= 1;
			if(j & 1) {
				fprintf(f, "dplonk(%d)", j >> 1);
			} else {
				j >>= 1;
				if(j & 1) {
					fprintf(f, "fadeup(%d)", j >> 1);
				} else {
					fprintf(f, "fadedn(%d)", j >> 1);
				}
			}
		}
	}
}

void enumerate(uint8_t n) {
	static uint8_t next = 0;

	node[n].address = next++;
	if(node[n].type == N_BRANCH) {
		(void) enumerate(node[n].alt[0]);
		(void) enumerate(node[n].alt[1]);
		node[n].value = BRANCH(node[node[n].alt[1]].address - (node[n].address + 1));
	}
}

int main() {
	char buf[32], str[32];
	int i, j, p;
	int lastnote = 16;
	int root;

	memset(leaves, 0xff, 256);

	while(!feof(stdin) && fgets(buf, sizeof(buf), stdin)) {
		if(1 == sscanf(buf, "p %s", str)) {
			for(i = 0; i < MAXNOTE; i++) {
				if(!strcmp(notemap[i].name, str)) {
					break;
				}
			}
			if(i == MAXNOTE) errx(1, "Unknown note");
			sequence[nseq++] = PLONK((notemap[i].id - lastnote) & 31);
			lastnote = notemap[i].id;
			if(!fgets(buf, sizeof(buf), stdin)) errx(1, "Unexpected EOF");
			if(1 != sscanf(buf, "t %d", &p) || p < PLONKTIME) {
				errx(1, "Plonk must be followed by t %d", PLONKTIME);
			}
			p -= PLONKTIME;
			while(p >= CHUNKTIME1) {
				sequence[nseq++] = DELAY(CHUNKTIME1);
				p -= CHUNKTIME1;
			}
			while(p >= CHUNKTIME2) {
				sequence[nseq++] = DELAY(CHUNKTIME2);
				p -= CHUNKTIME2;
			}
			/*while(p >= CHUNKTIME3) {
				sequence[nseq++] = DELAY(CHUNKTIME3);
				p -= CHUNKTIME3;
			}*/
			if(p) sequence[nseq++] = DELAY(p);
		} else if(1 == sscanf(buf, "t %d", &p)) {
			while(p >= CHUNKTIME1) {
				sequence[nseq++] = DELAY(CHUNKTIME1);
				p -= CHUNKTIME1;
			}
			while(p >= CHUNKTIME2) {
				sequence[nseq++] = DELAY(CHUNKTIME2);
				p -= CHUNKTIME2;
			}
			/*while(p >= CHUNKTIME3) {
				sequence[nseq++] = DELAY(CHUNKTIME3);
				p -= CHUNKTIME3;
			}*/
			if(p) sequence[nseq++] = DELAY(p);
		//} else if(1 == sscanf(buf, "v %d", &p)) {
		//	sequence[nseq++] = VOL(p / 2);
		} else if(1 == sscanf(buf, "f %d", &p)) {
			if(p >= 0) {
				sequence[nseq++] = FADEUP(p);
			} else {
				sequence[nseq++] = FADEDN(-p);
			}
		}
	}
	//sequence[nseq++] = STOP;

	for(i = 0; i < nseq; i++) {
		if(leaves[sequence[i]] == 0xff) {
			leaves[sequence[i]] = nnode;
			node[nnode].type = N_LEAF;
			node[nnode].value = sequence[i];
			node[nnode].weight = 1;
			nnode++;
		} else {
			node[leaves[sequence[i]]].weight++;
		}
	}

	for(;;) {
		node[nnode].type = N_BRANCH;
		i = lightest();
		node[nnode].alt[0] = i;
		node[nnode].weight = node[i].weight;
		node[i].origweight = node[i].weight;
		node[i].weight = 0;
		j = lightest();
		if(j == -1) break;
		node[nnode].alt[1] = j;
		node[nnode].weight += node[j].weight;
		node[j].origweight = node[j].weight;
		node[j].weight = 0;
		node[i].parent = nnode;
		node[j].parent = nnode;
		nnode++;
	}
	node[i].parent = 0xff;
	root = i;

#if 0
	for(i = 0; i < nnode; i++) {
		if(node[i].type == N_LEAF) {
			printf("%3d: %3d L $%02x ", i, node[i].origweight, node[i].value);
			fprintaction(stdout, node[i].value);
			printf("\n");
		} else if(node[i].type == N_BRANCH) {
			printf("%3d: %3d B %d %d\n", i, node[i].origweight, node[i].alt[0], node[i].alt[1]);
		}
	}
#endif

	enumerate(root);
	printf("tree:\n");
	for(i = 0; i < nnode; i++) {
		for(j = 0; j < nnode; j++) {
			if(node[j].address == i) {
				printf("\t\t.byte\t0x%02x\t; ", node[j].value);
				fprintaction(stdout, node[j].value);
				printf("\n");
				totalbytes++;
			}
		}
	}

	printf("bitstream:\n");
	j = 16;
	for(i = 0; i < nseq; i++) {
		fprintaction(stderr, sequence[i]);
		if((sequence[i] & 7) == 4) {
			j += sequence[i] >> 3;
			j &= 31;
			fprintf(stderr, " - %d", j);
		}
		fprintf(stderr, "\n");
		chain(leaves[sequence[i]]);
	}

	while(bitcount && accum) putbit(0, 0);

	fprintf(stderr, "%d bytes\n", totalbytes);
	return 0;
}
