SREG		= 0x3f
GIMSK		= 0x3b
GIFR		= 0x3a
TIMSK		= 0x39
TIFR		= 0x38
MCUCR		= 0x35
MCUSR		= 0x34
TCCR0		= 0x33
TCNT0		= 0x32
OSCCAL		= 0x31
TCCR1		= 0x30
TCNT1		= 0x2f
OCR1A		= 0x2e
OCR1B		= 0x2d
SFIOR		= 0x2c
WDTCR		= 0x21
EEAR		= 0x1e
EEDR		= 0x1d
EECR		= 0x1c
PORTB		= 0x18
DDRB		= 0x17
PINB		= 0x16
ACSR		= 0x08
ADMUX		= 0x07

VRATE		= 16

PLONKTIME	= 18

;	r0				Working register

;	r1, r2, r3, r4, r5, r6	free

;	r7				Volume delta

;	r8, r9				Oscillator phase
;	r10, r11			Oscillator frequency
;	r12				Oscillator volume [0, 15]
;	r13				Saved status register

;	r14, r15		free

;	r16, r17, r18, r19		Work registers

;	r20				Current note [0, 31]

;	r21				Delay timer

;	r22				Vibrato depth, should have 1 bit set (0x80 = none, 0x01 = full)
;	r23				Vibrato phase

;	r24				Next bitstream address
;	r25				Bitstream reader bit accumulator

;	r26				Work register in interrupt context

;	r27				Vibrato timer
;	r28				Volume [0, 255]

;	r29			free

;	r30, r31			Z: Work register, ROM pointer, r31 is always 0

		// 20 notes and two vectors

		rjmp	vec_init
		.byte	0x52		; e-1
		.byte	0xa2		; f#1
		.byte	0xc2		; g-1*
		.byte	0xf2		; g#1
		.byte	0x23		; a-1
		.byte	0x53		; a#1
		rjmp	vec_ovf1
		.byte	0x83		; h-1
		.byte	0xb3		; c-2
		.byte	0xf3		; c#2
		.byte	0x24		; d-2
		.byte	0xb4		; e-2
		.byte	0xf4		; f-2
		.byte	0x45		; f#2
		.byte	0x95		; g-2
		.byte	0xe5		; g#2
		.byte	0x46		; a-2
		.byte	0xa6		; a#2
		.byte	0x07		; h-2
		.byte	0x77		; c-3
		.byte	0x58		; d-3

#include "song.i"

		.align	2

vec_init:
		eor	r7, r7
		ldi	r20, 16
		ldi	r21, 0xff
		ldi	r24, lo8(bitstream)
		ldi	r25, 0x01
		ldi	r28, 0
		ldi	r31, 0

		ldi	r16, 0x1f
		out	DDRB, r16

		ldi	r16, 0x04
		out	TIMSK, r16
		ldi	r16, 0x03
		out	TCCR1, r16
		out	TCCR0, r16

loop:
		subi	r21, 1
		breq	parse

update:
		; Note to freq

		mov	r30, r20
		lpm
		mov	r16, r0
		mov	r17, r0
		andi	r16, 0xf0
		andi	r17, 0x0f

		; Add vibrato

		mov	r18, r23
		sbrc	r18, 7
		com	r18		; r18 in [0, 127]
		subi	r18, 64		; r18 in [-64, 63]

		mov	r19, r22
2:
		asr	r18
		lsr	r19
		brne	2b
		; r19 is 0x00

		sbrc	r18, 7
		ldi	r19, 0xff

		add	r16, r18
		adc	r17, r19

		; Extend volume

		mov	r18, r28
		swap	r18
		andi	r18, 15

		; Update oscillator inputs

		cli
		mov	r10, r16
		mov	r11, r17
		mov	r12, r18
		sei

		; Animate vibrato

		subi	r23, VRATE

		subi	r27, 0x01
		brcc	4f

		ldi	r27, 16
		sbrs	r22, 0
		lsr	r22
4:
		; Animate volume

		add	r28, r7
		sbrc	r7, 7
		rjmp	2f

		brcc	1f
		ser	r28
		rjmp	1f
2:
		brcs	1f
		clr	r28
1:
		; sync to control rate timer

3:		in	r16, TIFR
		sbrs	r16, 1
		rjmp	3b
		ldi	r16, 0x02
		out	TIFR, r16

		rjmp	loop

tok_fade:
		lsr	r0
		brcs	tok_fadeup
tok_fadedn:
		neg	r0
tok_fadeup:
		mov	r7, r0

parse:
		; Read command

		ldi	r30, lo8(tree)

gettoken:
		lpm
		lsr	r0
		brcs	tok_branch
		lsr	r0
		brcs	tok_delay
		lsr	r0
		brcc	tok_fade

tok_plonk:
		add	r20, r0
		andi	r20, 31
		ldi	r22, 8
		ldi	r27, 116
		ldi	r21, PLONKTIME*2
		rjmp	update

tok_branch:
		mov	r17, r0

		lsl	r25
		brne	1f

		mov	r16, r30

		mov	r30, r24
		tst	r30
		breq	eof

		lpm
		mov	r25, r0
		inc	r24
eof:
		mov	r30, r16

		sec
		rol	r25
1:
		brcc	2f
		add	r30, r17
2:
		inc	r30

		rjmp	gettoken

tok_delay:
		mov	r21, r0
		lsl	r21
		rjmp	update

vec_ovf1:
		in	r13, SREG
		add	r8, r10
		adc	r9, r11
		mov	r26, r12
		sbrc	r9, 7
		neg	r26
		subi	r26, 256-16
		out	PORTB, r26
		out	SREG, r13
		reti

		.byte	'L', 'F', 'T', '!'
